; iconv.def (emx+gcc)
LIBRARY ICONV 
DESCRIPTION 'OS/2 port of libiconv <http://clisp.cons.org/~haible/packages-libiconv.html>'
EXPORTS
	libiconv_open	@1
	libiconv	@2
	libiconv_close	@3
	libiconvctl	@4
